@@ -1,20 +1,27 @@
 """
 Configuration object for the Tribler Core.
 """
+from __future__ import absolute_import
+
 import logging
 import os
 
 from configobj import ConfigObj
+
+from six import text_type
+
 from validate import Validator
 
+from Tribler.Core import version as tribler_version
+from Tribler.Core.Config.download_config import get_default_dest_dir
 from Tribler.Core.Utilities.install_dir import get_lib_path
 from Tribler.Core.Utilities.network_utils import get_random_port
+from Tribler.Core.Utilities.unicode import ensure_unicode
 from Tribler.Core.exceptions import InvalidConfigException
 from Tribler.Core.osutils import get_appstate_dir
-from Tribler.Core.DownloadConfig import get_default_dest_dir
 
-FILENAME = 'triblerd.conf'
-SPEC_FILENAME = 'config.spec'
+CONFIG_FILENAME = 'triblerd.conf'
+SPEC_FILENAME = 'tribler_config.spec'
 CONFIG_SPEC_PATH = os.path.join(get_lib_path(), 'Core', 'Config', SPEC_FILENAME)
 
 
@@ -35,12 +42,13 @@ class TriblerConfig(object):
         """
         self._logger = logging.getLogger(self.__class__.__name__)
 
+        self._root_state_dir = self.get_default_root_state_dir()
+        self._state_dir = os.path.join(self._root_state_dir, tribler_version.version_id)
+
         if config is None:
-            file_name = os.path.join(self.get_default_state_dir(), FILENAME)
-            if os.path.exists(file_name):
-                config = ConfigObj(file_name, configspec=CONFIG_SPEC_PATH, encoding='latin_1')
-            else:
-                config = ConfigObj(configspec=CONFIG_SPEC_PATH, encoding='latin_1')
+            config_file = os.path.join(self._state_dir, CONFIG_FILENAME)
+            config = ConfigObj(infile=(config_file if os.path.exists(config_file) else None),
+                               configspec=CONFIG_SPEC_PATH, default_encoding='utf-8')
         self.config = config
         self.validate()
 
@@ -49,19 +57,32 @@ class TriblerConfig(object):
             self.config['download_defaults']['saveas'] = get_default_dest_dir()
         self.selected_ports = {}
 
+    def abspath(self, path):
+        return path if os.path.isabs(path) else os.path.join(self.get_state_dir(), path)
+
+    def norm_path(self, path):
+        """
+        Return absolute path if it points outside the state dir. Otherwise, change it to relative path.
+        """
+        if os.path.isabs(path):
+            base_path = self.get_state_dir()
+            if base_path == os.path.commonprefix([path, base_path]):
+                return os.path.relpath(path, base_path)
+        return path
+
     @staticmethod
     def load(config_path=None):
         """
         Load a TriblerConfig from disk.
         """
-        return TriblerConfig(ConfigObj(config_path, configspec=CONFIG_SPEC_PATH, encoding='latin_1'))
+        return TriblerConfig(ConfigObj(config_path, configspec=CONFIG_SPEC_PATH, default_encoding='utf-8'))
 
     def copy(self):
         """
         Return a TriblerConfig object that has the same values.
         """
         # References to the sections are copied here
-        new_configobj = ConfigObj(self.config.copy(), configspec=self.config.configspec, encoding='latin_1')
+        new_configobj = ConfigObj(self.config.copy(), configspec=self.config.configspec, default_encoding='utf-8')
         # Make a deep copy of every section
         for section in self.config:
             new_configobj[section] = self.config[section].copy()
@@ -85,20 +106,18 @@ class TriblerConfig(object):
         """
         if not os.path.exists(self.get_state_dir()):
             os.makedirs(self.get_state_dir())
-        with open(os.path.join(self.get_state_dir(), FILENAME), 'w') as outfile:
-            self.config.write(outfile=outfile)
+        self.update_version()
+        self.config.filename = os.path.join(self.get_state_dir(), CONFIG_FILENAME)
+        self.config.write()
 
     @staticmethod
-    def get_default_state_dir(home_dir_postfix=u'.Tribler'):
+    def get_default_root_state_dir(home_dir_postfix=u'.Tribler'):
         """Get the default application state directory."""
         if 'TSTATEDIR' in os.environ:
-            return os.environ['TSTATEDIR']
+            path = os.environ['TSTATEDIR']
+            return path if os.path.isabs(path) else os.path.join(os.getcwd(), path)
 
-        if os.path.isdir(home_dir_postfix):
-            return os.path.abspath(home_dir_postfix)
-
-        application_directory = get_appstate_dir()
-        return os.path.join(application_directory, home_dir_postfix)
+        return os.path.join(get_appstate_dir(), home_dir_postfix)
 
     def _obtain_port(self, section, option):
         """
@@ -120,52 +139,60 @@ class TriblerConfig(object):
             self._logger.debug(u"Get random port %d for [%s]", self.selected_ports[path], path)
         return self.selected_ports[path]
 
-    # General
+    # Version and backup
+    def set_version(self, version):
+        self.config['general']['version'] = version
 
-    def set_family_filter_enabled(self, value):
-        self.config['general']['family_filter'] = bool(value)
+    def get_version(self):
+        return self.config['general']['version']
 
-    def get_family_filter_enabled(self):
-        return self.config['general'].as_bool('family_filter')
+    def update_version(self):
+        self.config['general']['version'] = tribler_version.version_id
 
-    def set_state_dir(self, state_dir):
-        self.config["general"]["state_dir"] = state_dir
+    def set_version_backup_enabled(self, backup_enabled):
+        self.config['general']['version_backup_enabled'] = backup_enabled
 
-    def get_state_dir(self):
-        if not self.config["general"]["state_dir"]:
-            self.set_state_dir(TriblerConfig.get_default_state_dir())
+    def get_version_backup_enabled(self):
+        return self.config['general']['version_backup_enabled']
 
-        return self.config["general"]["state_dir"]
+    # Chant
+    def set_chant_enabled(self, value):
+        self.config['chant']['enabled'] = bool(value)
 
-    def set_permid_keypair_filename(self, keypair_filename):
-        self.config['general']['ec_keypair_filename'] = keypair_filename
+    def get_chant_enabled(self):
+        return self.config['chant']['enabled']
 
-    def get_permid_keypair_filename(self):
-        file_name = self.config["general"]["ec_keypair_filename"]
-        if not file_name:
-            file_name = os.path.join(self.get_state_dir(), 'ec.pem')
-            self.set_permid_keypair_filename(file_name)
-        return file_name
+    def set_chant_channels_dir(self, chant_db_filename):
+        self.config['chant']['channels_dir'] = self.norm_path(chant_db_filename)
+
+    def get_chant_channels_dir(self):
+        return self.abspath(self.config['chant']['channels_dir'])
+
+    def set_root_state_dir(self, root_dir, version_id=None):
+        """
+        Sets the root state directory for Tribler. The versioning is handled automatically.
+        """
+        self._root_state_dir = root_dir
+        self._state_dir = os.path.join(self._root_state_dir, version_id or tribler_version.version_id)
+
+    def get_root_state_dir(self):
+        return self._root_state_dir
+
+    def get_state_dir(self, version_id=None):
+        return os.path.join(self._root_state_dir, version_id) if version_id \
+            else ensure_unicode(self._state_dir, 'utf-8')
 
     def set_trustchain_keypair_filename(self, keypairfilename):
-        self.config['trustchain']['ec_keypair_filename'] = keypairfilename
+        self.config['trustchain']['ec_keypair_filename'] = self.norm_path(keypairfilename)
 
     def get_trustchain_keypair_filename(self):
-        file_name = self.config['trustchain']['ec_keypair_filename']
-        if not file_name:
-            file_name = os.path.join(self.get_state_dir(), 'ec_multichain.pem')
-            self.set_trustchain_keypair_filename(file_name)
-        return file_name
+        return self.abspath(self.config['trustchain']['ec_keypair_filename'])
 
     def set_trustchain_testnet_keypair_filename(self, keypairfilename):
-        self.config['trustchain']['testnet_keypair_filename'] = keypairfilename
+        self.config['trustchain']['testnet_keypair_filename'] = self.norm_path(keypairfilename)
 
     def get_trustchain_testnet_keypair_filename(self):
-        file_name = self.config['trustchain']['testnet_keypair_filename']
-        if not file_name:
-            file_name = os.path.join(self.get_state_dir(), 'ec_trustchain_testnet.pem')
-            self.set_trustchain_testnet_keypair_filename(file_name)
-        return file_name
+        return self.abspath(self.config['trustchain']['testnet_keypair_filename'])
 
     def set_trustchain_enabled(self, value):
         self.config['trustchain']['enabled'] = value
@@ -176,21 +203,32 @@ class TriblerConfig(object):
     def set_trustchain_live_edges_enabled(self, value):
         self.config['trustchain']['live_edges_enabled'] = value
 
+    def set_bootstrap_enabled(self, value):
+        self.config['bootstrap']['enabled'] = value
+
+    def get_bootstrap_enabled(self):
+        return self.config['bootstrap']['enabled']
+
+    def get_bootstrap_max_download_rate(self):
+        return self.config['bootstrap'].as_int('max_download_rate')
+
+    def set_bootstrap_max_download_rate(self, value):
+        self.config['bootstrap']['max_download_rate'] = value
+
+    def get_bootstrap_infohash(self):
+        return self.config['bootstrap']['infohash']
+
+    def set_bootstrap_infohash(self, value):
+        self.config['bootstrap']['infohash'] = value
+
     def get_trustchain_live_edges_enabled(self):
         return self.config['trustchain']['live_edges_enabled']
 
-    def set_megacache_enabled(self, value):
-        self.config['general']['megacache'] = value
-
-    def get_megacache_enabled(self):
-        return self.config['general']['megacache']
-
     def set_log_dir(self, value):
-        self.config['general']['log_dir'] = value
+        self.config['general']['log_dir'] = self.norm_path(value)
 
     def get_log_dir(self):
-        log_dir = self.config['general']['log_dir']
-        return os.path.join(self.get_state_dir(), 'logs') if (not log_dir or log_dir == 'None') else log_dir
+        return self.abspath(self.config['general']['log_dir'])
 
     def set_testnet(self, value):
         self.config['general']['testnet'] = value
@@ -235,20 +273,6 @@ class TriblerConfig(object):
     def get_http_api_retry_port(self):
         return self.config['http_api']['retry_port']
 
-    # Dispersy
-
-    def set_dispersy_enabled(self, value):
-        self.config['dispersy']['enabled'] = value
-
-    def get_dispersy_enabled(self):
-        return self.config['dispersy']['enabled']
-
-    def set_dispersy_port(self, value):
-        self.config['dispersy']['port'] = value
-
-    def get_dispersy_port(self):
-        return self._obtain_port('dispersy', 'port')
-
     # IPv8
 
     def set_ipv8_enabled(self, value):
@@ -257,6 +281,12 @@ class TriblerConfig(object):
     def get_ipv8_enabled(self):
         return self.config['ipv8']['enabled']
 
+    def set_ipv8_port(self, value):
+        self.config['ipv8']['port'] = value
+
+    def get_ipv8_port(self):
+        return self._obtain_port('ipv8', 'port')
+
     def set_ipv8_bootstrap_override(self, value):
         self.config['ipv8']['bootstrap_override'] = value
 
@@ -424,20 +454,6 @@ class TriblerConfig(object):
     def get_libtorrent_dht_enabled(self):
         return self.config['libtorrent']['dht']
 
-    # Mainline DHT
-
-    def set_mainline_dht_enabled(self, value):
-        self.config['mainline_dht']['enabled'] = value
-
-    def get_mainline_dht_enabled(self):
-        return self.config['mainline_dht']['enabled']
-
-    def set_mainline_dht_port(self, port):
-        self.config['mainline_dht']['port'] = port
-
-    def get_mainline_dht_port(self):
-        return self._obtain_port('mainline_dht', 'port')
-
     # Video server
 
     def set_video_server_enabled(self, value):
@@ -466,7 +482,7 @@ class TriblerConfig(object):
     def get_tunnel_community_socks5_listen_ports(self):
         ports = self.config['tunnel_community']['socks5_listen_ports']
         path = u'tunnel_community~socks5_listen_ports~'
-        return [self._get_random_port(path + unicode(index))
+        return [self._get_random_port(path + text_type(index))
                 if int(port) < 0 else int(port) for index, port in enumerate(ports)]
 
     def set_tunnel_community_exitnode_enabled(self, value):
@@ -506,10 +522,34 @@ class TriblerConfig(object):
         return self.config['download_defaults']['safeseeding_enabled']
 
     def set_default_destination_dir(self, value):
-        self.config['download_defaults']['saveas'] = value
+        self.config['download_defaults']['saveas'] = self.norm_path(value)
 
     def get_default_destination_dir(self):
-        return self.config['download_defaults']['saveas']
+        return self.abspath(self.config['download_defaults']['saveas'])
+
+    def set_default_add_download_to_channel(self, value):
+        self.config['download_defaults']['add_download_to_channel'] = value
+
+    def get_default_add_download_to_channel(self):
+        return self.config['download_defaults']['add_download_to_channel']
+
+    def set_seeding_mode(self, value):
+        self.config['download_defaults']['seeding_mode'] = value
+
+    def get_seeding_mode(self):
+        return self.config['download_defaults']['seeding_mode']
+
+    def set_seeding_time(self, value):
+        self.config['download_defaults']['seeding_time'] = value
+
+    def get_seeding_time(self):
+        return self.config['download_defaults']['seeding_time']
+
+    def set_seeding_ratio(self, value):
+        self.config['download_defaults']['seeding_ratio'] = value
+
+    def get_seeding_ratio(self):
+        return self.config['download_defaults']['seeding_ratio']
 
     # Market Community
 
@@ -561,86 +601,6 @@ class TriblerConfig(object):
     def set_popularity_community_enabled(self, value):
         self.config['popularity_community']['enabled'] = value
 
-    # Torrent store
-
-    def get_torrent_store_enabled(self):
-        return self.config['torrent_store']['enabled']
-
-    def set_torrent_store_enabled(self, value):
-        self.config['torrent_store']['enabled'] = value
-
-    def get_torrent_store_dir(self):
-        return os.path.join(self.get_state_dir(), self.config['torrent_store']['store_dir'])
-
-    def set_torrent_store_dir(self, value):
-        self.config['torrent_store']['store_dir'] = value
-
-    # Metadata
-
-    def get_metadata_enabled(self):
-        return self.config['metadata']['enabled']
-
-    def set_metadata_enabled(self, mode):
-        self.config['metadata']['enabled'] = mode
-
-    def get_metadata_store_dir(self):
-        return os.path.join(self.get_state_dir(), self.config['metadata']['store_dir'])
-
-    def set_metadata_store_dir(self, value):
-        self.config['metadata']['store_dir'] = value
-
-    # Torrent collecting
-
-    def set_torrent_collecting_enabled(self, value):
-        self.config['torrent_collecting']['enabled'] = value
-
-    def get_torrent_collecting_enabled(self):
-        return self.config['torrent_collecting']['enabled']
-
-    def set_torrent_collecting_max_torrents(self, value):
-        self.config['torrent_collecting']['max_torrents'] = value
-
-    def get_torrent_collecting_max_torrents(self):
-        return self.config['torrent_collecting']['max_torrents']
-
-    def set_torrent_collecting_dir(self, value):
-        self.config['torrent_collecting']['directory'] = value
-
-    def get_torrent_collecting_dir(self):
-        return self.config['torrent_collecting']['directory']
-
-    # Search Community
-
-    def set_torrent_search_enabled(self, mode):
-        self.config['search_community']['enabled'] = mode
-
-    def get_torrent_search_enabled(self):
-        return self.config['search_community']['enabled']
-
-    # AllChannel Community
-
-    def set_channel_search_enabled(self, mode):
-        self.config['allchannel_community']['enabled'] = mode
-
-    def get_channel_search_enabled(self):
-        return self.config['allchannel_community']['enabled']
-
-    # Channel Community
-
-    def set_channel_community_enabled(self, value):
-        self.config['channel_community']['enabled'] = value
-
-    def get_channel_community_enabled(self):
-        return self.config['channel_community']['enabled']
-
-    # PreviewChannel Community
-
-    def set_preview_channel_community_enabled(self, value):
-        self.config['preview_channel_community']['enabled'] = value
-
-    def get_preview_channel_community_enabled(self):
-        return self.config['preview_channel_community']['enabled']
-
     # Watch folder
 
     def set_watch_folder_enabled(self, value):
@@ -650,10 +610,10 @@ class TriblerConfig(object):
         return self.config['watch_folder']['enabled']
 
     def set_watch_folder_path(self, value):
-        self.config['watch_folder']['directory'] = value
+        self.config['watch_folder']['directory'] = self.norm_path(value)
 
     def get_watch_folder_path(self):
-        return self.config['watch_folder']['directory']
+        return self.abspath(self.config['watch_folder']['directory'])
 
     # Resource monitor
 
